@@ -10,17 +10,15 @@ mod raw {
     use std::cmp::min;
     use std::fs::File;
     use std::io::{self, Read, Write};
-    use std::os::unix::io::AsRawFd;
     use std::time::{Duration, Instant};
 
-    fn as_pollfd(f: Option<&File>, for_read: bool) -> posix::PollFd {
-        let optfd = f.map(File::as_raw_fd);
+    fn as_pollfd<'a>(f: Option<&'a File>, for_read: bool) -> posix::PollFd<'a> {
         let events = if for_read {
             posix::POLLIN
         } else {
             posix::POLLOUT
         };
-        posix::PollFd::new(optfd, events)
+        posix::PollFd::new(f, events)
     }
 
     fn maybe_poll(
@@ -155,8 +153,7 @@ mod raw {
                         // close stdin when done writing, so the child receives EOF
                         self.stdin.take();
                         // deallocate the input data, we don't need it any more
-                        self.input_data.clear();
-                        self.input_data.shrink_to_fit();
+                        self.input_data = Vec::new();
                     }
                 }
                 if out_ready {
@@ -185,18 +182,17 @@ mod raw {
             deadline: Option<Instant>,
             size_limit: Option<usize>,
         ) -> (Option<io::Error>, (Option<Vec<u8>>, Option<Vec<u8>>)) {
-            let mut outvec = Vec::<u8>::new();
-            let mut errvec = Vec::<u8>::new();
+            let mut outvec = vec![];
+            let mut errvec = vec![];
 
-            let check = self.read_into(deadline, size_limit, &mut outvec, &mut errvec);
+            let err = self
+                .read_into(deadline, size_limit, &mut outvec, &mut errvec)
+                .err();
             let output = (
                 self.stdout.as_ref().map(|_| outvec),
                 self.stderr.as_ref().map(|_| errvec),
             );
-            match check {
-                Ok(()) => (None, output),
-                Err(e) => (Some(e), output),
-            }
+            (err, output)
         }
     }
 }
@@ -205,7 +201,6 @@ mod raw {
 mod raw {
     use std::fs::File;
     use std::io::{self, Read, Write};
-    use std::mem;
     use std::sync::mpsc::{self, RecvTimeoutError, SyncSender};
     use std::thread;
     use std::time::Instant;
@@ -251,7 +246,7 @@ mod raw {
         }
     }
 
-    fn spawn_curried<T: Send + 'static>(f: impl FnOnce(T) + Send + 'static, arg: T) {
+    fn spawn_with_arg<T: Send + 'static>(f: impl FnOnce(T) + Send + 'static, arg: T) {
         thread::spawn(move || f(arg));
     }
 
@@ -289,16 +284,16 @@ mod raw {
                 let input_data = input_data.expect("must provide input to redirected stdin");
                 helper_set |= StreamIdent::In as u8;
                 move |tx: SyncSender<_>| match stdin.write_all(&input_data) {
-                    Ok(()) => mem::drop(tx.send((StreamIdent::In, Payload::EOF))),
-                    Err(e) => mem::drop(tx.send((StreamIdent::In, Payload::Err(e)))),
+                    Ok(()) => drop(tx.send((StreamIdent::In, Payload::EOF))),
+                    Err(e) => drop(tx.send((StreamIdent::In, Payload::Err(e)))),
                 }
             });
 
             let (tx, rx) = mpsc::sync_channel(0);
 
-            read_stdout.map(|f| spawn_curried(f, tx.clone()));
-            read_stderr.map(|f| spawn_curried(f, tx.clone()));
-            write_stdin.map(|f| spawn_curried(f, tx.clone()));
+            read_stdout.map(|f| spawn_with_arg(f, tx.clone()));
+            read_stderr.map(|f| spawn_with_arg(f, tx.clone()));
+            write_stdin.map(|f| spawn_with_arg(f, tx.clone()));
 
             RawCommunicator {
                 rx,
@@ -310,11 +305,10 @@ mod raw {
 
         fn recv_until(&self, deadline: Option<Instant>) -> Result<Message, Timeout> {
             if let Some(deadline) = deadline {
-                let now = Instant::now();
-                if now >= deadline {
-                    return Err(Timeout);
-                }
-                match self.rx.recv_timeout(deadline - now) {
+                match self
+                    .rx
+                    .recv_timeout(deadline.saturating_duration_since(Instant::now()))
+                {
                     Ok(message) => Ok(message),
                     Err(RecvTimeoutError::Timeout) => Err(Timeout),
                     // should never be disconnected, the helper threads always
@@ -398,7 +392,9 @@ mod raw {
             let mut outvec = vec![];
             let mut errvec = vec![];
 
-            let check = self.read_into(deadline, size_limit, &mut outvec, &mut errvec);
+            let err = self
+                .read_into(deadline, size_limit, &mut outvec, &mut errvec)
+                .err();
             let output = {
                 let (mut o, mut e) = (None, None);
                 if self.requested_streams & StreamIdent::Out as u8 != 0 {
@@ -413,23 +409,24 @@ mod raw {
                 }
                 (o, e)
             };
-            match check {
-                Ok(()) => (None, output),
-                Err(e) => (Some(e), output),
-            }
+            (err, output)
         }
     }
 }
 
 use raw::RawCommunicator;
 
-/// Deadlock-free communication with the subprocess.
+/// Unattended data exchange with the subprocess.
 ///
-/// Normally care must be taken to avoid deadlock when communicating to a
-/// subprocess that both expects input and provides output.  This
-/// implementation avoids deadlock by reading from and writing to the
+/// When a subprocess both expects input and provides output, care must be
+/// taken to avoid deadlock.  The issue arises when the subprocess responds to
+/// part of the input data by providing some output which must be read for the
+/// subprocess to accept further input.  If the parent process is blocked on
+/// writing the input, it cannot read the output and a deadlock occurs.  This
+/// implementation avoids this issue by by reading from and writing to the
 /// subprocess in parallel.  On Unix-like systems this is achieved using
 /// `poll()`, and on Windows using threads.
+#[must_use]
 #[derive(Debug)]
 pub struct Communicator {
     inner: RawCommunicator,
@@ -514,10 +511,7 @@ impl Communicator {
     /// replacement character.
     pub fn read_string(&mut self) -> Result<(Option<String>, Option<String>), CommunicateError> {
         let (o, e) = self.read()?;
-        Ok((
-            o.map(|v| String::from_utf8_lossy(&v).into()),
-            e.map(|v| String::from_utf8_lossy(&v).into()),
-        ))
+        Ok((o.map(from_utf8_lossy), e.map(from_utf8_lossy)))
     }
 
     /// Limit the amount of data the next `read()` will read from the
@@ -535,6 +529,15 @@ impl Communicator {
     }
 }
 
+/// Like String::from_utf8_lossy(), but takes `Vec<u8>` and reuses its storage if
+/// possible.
+fn from_utf8_lossy(v: Vec<u8>) -> String {
+    match String::from_utf8(v) {
+        Ok(s) => s,
+        Err(e) => String::from_utf8_lossy(e.as_bytes()).into(),
+    }
+}
+
 pub fn communicate(
     stdin: Option<File>,
     stdout: Option<File>,
@@ -579,17 +582,13 @@ impl CommunicateError {
 }
 
 impl Error for CommunicateError {
-    fn description(&self) -> &str {
-        self.error.description()
-    }
-
-    fn cause(&self) -> Option<&dyn Error> {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
         self.error.source()
     }
 }
 
 impl fmt::Display for CommunicateError {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.error.fmt(f)
     }
 }
