@@ -1,3 +1,4 @@
+use std::cell::RefCell;
 use std::env;
 use std::error::Error;
 use std::ffi::{OsStr, OsString};
@@ -9,7 +10,7 @@ use std::result;
 use std::time::Duration;
 
 use crate::communicate;
-use crate::os_common::{self, ExitStatus, StandardStream};
+use crate::os_common::{ExitStatus, StandardStream};
 
 use self::ChildState::*;
 
@@ -30,15 +31,15 @@ pub use communicate::Communicator;
 /// error streams can be connected to the parent and available as [`stdin`],
 /// [`stdout`], and [`stderr`] public fields.  If you need to read the output
 /// and errors into memory (or provide input as a memory slice), use the
-/// [`communicate`] family of methods, which guarantee deadlock-free
-/// communication with the subprocess.
+/// [`communicate`] family of methods.
 ///
 /// `Popen` instances can be obtained with the [`create`] method, or
-/// using the [`popen`] method of the [`Exec`] class.  Subprocesses
+/// using the [`popen`] method of the [`Exec`] type.  Subprocesses
 /// can be connected into pipes, most easily achieved using using
 /// [`Exec`].
 ///
 /// [`Exec`]: struct.Exec.html
+/// [`popen`]: struct.Exec.html#method.popen
 /// [`stdin`]: struct.Popen.html#structfield.stdin
 /// [`stdout`]: struct.Popen.html#structfield.stdout
 /// [`stderr`]: struct.Popen.html#structfield.stderr
@@ -70,7 +71,11 @@ pub struct Popen {
 #[derive(Debug)]
 enum ChildState {
     Preparing, // only during construction
-    Running { pid: u32, ext: os::ExtChildState },
+    Running {
+        pid: u32,
+        #[allow(dead_code)]
+        ext: os::ExtChildState,
+    },
     Finished(ExitStatus),
 }
 
@@ -143,9 +148,19 @@ pub struct PopenConfig {
     /// Set group ID for the subprocess.
     ///
     /// If specified, calls `setgid()` before execing the child process.
+    ///
+    /// Not to be confused with similarly named `setpgid`.
     #[cfg(unix)]
     pub setgid: Option<u32>,
 
+    /// Make the subprocess belong to a new process group.
+    ///
+    /// If specified, calls `setpgid(0, 0)` before execing the child process.
+    ///
+    /// Not to be confused with similarly named `setgid`.
+    #[cfg(unix)]
+    pub setpgid: bool,
+
     // Add this field to force construction using ..Default::default() for
     // backward compatibility.  Unfortunately we can't mark this non-public
     // because then ..Default::default() wouldn't work either.
@@ -176,6 +191,8 @@ impl PopenConfig {
             setuid: self.setuid,
             #[cfg(unix)]
             setgid: self.setgid,
+            #[cfg(unix)]
+            setpgid: self.setpgid,
             _use_default_to_construct: (),
         })
     }
@@ -203,6 +220,8 @@ impl Default for PopenConfig {
             setuid: None,
             #[cfg(unix)]
             setgid: None,
+            #[cfg(unix)]
+            setpgid: false,
             _use_default_to_construct: (),
         }
     }
@@ -385,7 +404,7 @@ impl Popen {
             // the same File.  If the file is unavailable, use the
             // appropriate system output stream.
             if src.is_none() {
-                *src = Some(os_common::get_standard_stream(src_id)?);
+                *src = Some(get_standard_stream(src_id)?);
             }
             *dest = Some(Rc::clone(src.as_ref().unwrap()));
             Ok(())
@@ -486,13 +505,27 @@ impl Popen {
 
     /// Prepare to communicate with the subprocess.
     ///
-    /// Communicating refers to providing the given `input_data` to the
-    /// subprocess's standard input, while simultaneously reading from its
-    /// standard output and error streams.
+    /// Communicating refers to unattended data exchange with the subprocess.
+    /// During communication the given `input_data` is written to the
+    /// subprocess's standard input which is then closed, while simultaneously
+    /// its standard output and error streams are read until end-of-file is
+    /// reached.
     ///
-    /// This method does not start the actual communication, it just sets it
-    /// up and returns a [`Communicator`] handle.  Call the [`read`] method on
-    /// the returned handle to communicate.
+    /// The difference between this and simply writing input data to
+    /// `self.stdin` and then reading output from `self.stdout` and
+    /// `self.stderr` is that the reading and the writing are performed
+    /// simultaneously.  A naive implementation that writes and then reads has
+    /// an issue when the subprocess responds to part of the input by
+    /// providing output.  The output must be read for the subprocess to
+    /// accept further input, but the parent process is still blocked on
+    /// writing the rest of the input daata.  Since neither process can
+    /// proceed, a deadlock occurs.  This is why a correct implementation must
+    /// write and read at the same time.
+    ///
+    /// This method does not perform the actual communication, it just sets it
+    /// up and returns a [`Communicator`].  Call the [`read`] or
+    /// [`read_string`] method on the `Communicator` to exchange data with the
+    /// subprocess.
     ///
     /// Compared to `communicate()` and `communicate_bytes()`, the
     /// `Communicator` provides more control, such as timeout, read size
@@ -501,6 +534,7 @@ impl Popen {
     ///
     /// [`Communicator`]: struct.Communicator.html
     /// [`read`]: struct.Communicator.html#method.read
+    /// [`read_string`]: struct.Communicator.html#method.read_string
     pub fn communicate_start(&mut self, input_data: Option<Vec<u8>>) -> Communicator {
         communicate::communicate(
             self.stdin.take(),
@@ -518,8 +552,9 @@ impl Popen {
     /// The `None` options correspond to streams not specified as
     /// `Redirection::Pipe` when creating the subprocess.
     ///
-    /// The communication is deadlock-free, which is achieved by reading and
-    /// writing in parallel using `poll()` on Unix and threads on Windows.
+    /// This implementation reads and writes simultaneously, avoiding deadlock
+    /// in case the subprocess starts writing output before reading the whole
+    /// input - see [`communicate_start()`] for details.
     ///
     /// Note that this method does not wait for the subprocess to finish, only
     /// to close its output/error streams.  It is rare but possible for the
@@ -583,7 +618,7 @@ impl Popen {
     /// is still running, `None` is returned, otherwise
     /// `Some(exit_status)`.  This method is guaranteed not to block
     /// and is exactly equivalent to
-    /// `wait_timeout(Duration::new(0, 0)).unwrap_or(None)`.
+    /// `wait_timeout(Duration::from_secs(0)).unwrap_or(None)`.
     pub fn poll(&mut self) -> Option<ExitStatus> {
         self.wait_timeout(Duration::from_secs(0)).unwrap_or(None)
     }
@@ -657,7 +692,6 @@ mod os {
     use std::ffi::OsString;
     use std::fs::File;
     use std::io::{self, Read, Write};
-    use std::mem;
     use std::os::unix::io::AsRawFd;
     use std::time::{Duration, Instant};
 
@@ -673,13 +707,9 @@ mod os {
             set_inheritable(&exec_fail_pipe.1, false)?;
             {
                 let child_ends = self.setup_streams(config.stdin, config.stdout, config.stderr)?;
-                let child_env = config.env.map(format_env);
+                let child_env = config.env.as_deref().map(format_env);
                 let cmd_to_exec = config.executable.as_ref().unwrap_or(&argv[0]);
-                let just_exec = posix::stage_exec(
-                    cmd_to_exec,
-                    &argv[..],
-                    child_env.as_ref().map(::std::ops::Deref::deref),
-                )?;
+                let just_exec = posix::prep_exec(cmd_to_exec, &argv, child_env.as_deref())?;
                 unsafe {
                     // unsafe because after the call to fork() the
                     // child is not allowed to allocate
@@ -691,13 +721,14 @@ mod os {
                             };
                         }
                         None => {
-                            mem::drop(exec_fail_pipe.0);
+                            drop(exec_fail_pipe.0);
                             let result = Popen::do_exec(
                                 just_exec,
                                 child_ends,
-                                config.cwd.as_ref().map(|x| &x[..]),
+                                config.cwd.as_deref(),
                                 config.setuid,
                                 config.setgid,
+                                config.setpgid,
                             );
                             // If we are here, it means that exec has failed.  Notify
                             // the parent and exit.
@@ -719,7 +750,7 @@ mod os {
                     }
                 }
             }
-            mem::drop(exec_fail_pipe.1);
+            drop(exec_fail_pipe.1);
             let mut error_buf = [0u8; 4];
             let read_cnt = exec_fail_pipe.0.read(&mut error_buf)?;
             if read_cnt == 0 {
@@ -779,16 +810,15 @@ mod os {
         }
     }
 
-    fn format_env(env: Vec<(OsString, OsString)>) -> Vec<OsString> {
-        // Convert Vec of (key, val) pairs to Vec of key=val, as
-        // required by execvpe.  Also eliminate dups, with the
-        // later-appearing one taking precedence.
-        let mut seen = HashSet::<OsString>::new();
+    fn format_env(env: &[(OsString, OsString)]) -> Vec<OsString> {
+        // Convert Vec of (key, val) pairs to Vec of key=val, as required by
+        // execvpe.  Eliminate dups, in favor of later-appearing entries.
+        let mut seen = HashSet::<&OsStr>::new();
         let mut formatted: Vec<_> = env
-            .into_iter()
+            .iter()
             .rev()
-            .filter(|&(ref k, ref _v)| seen.insert(k.clone()))
-            .map(|(ref k, ref v)| {
+            .filter(|&(k, _)| seen.insert(k))
+            .map(|(k, v)| {
                 let mut fmt = k.clone();
                 fmt.push("=");
                 fmt.push(v);
@@ -801,22 +831,24 @@ mod os {
 
     trait PopenOsImpl: super::PopenOs {
         fn do_exec(
-            just_exec: impl Fn() -> io::Result<()>,
+            just_exec: impl FnOnce() -> io::Result<()>,
             child_ends: (Option<Rc<File>>, Option<Rc<File>>, Option<Rc<File>>),
             cwd: Option<&OsStr>,
             setuid: Option<u32>,
             setgid: Option<u32>,
+            setpgid: bool,
         ) -> io::Result<()>;
         fn waitpid(&mut self, block: bool) -> io::Result<()>;
     }
 
     impl PopenOsImpl for Popen {
         fn do_exec(
-            just_exec: impl Fn() -> io::Result<()>,
+            just_exec: impl FnOnce() -> io::Result<()>,
             child_ends: (Option<Rc<File>>, Option<Rc<File>>, Option<Rc<File>>),
             cwd: Option<&OsStr>,
             setuid: Option<u32>,
             setgid: Option<u32>,
+            setpgid: bool,
         ) -> io::Result<()> {
             if let Some(cwd) = cwd {
                 env::set_current_dir(cwd)?;
@@ -846,6 +878,9 @@ mod os {
             if let Some(gid) = setgid {
                 posix::setgid(gid)?;
             }
+            if setpgid {
+                posix::setpgid(0, 0)?;
+            }
             just_exec()?;
             unreachable!();
         }
@@ -965,7 +1000,7 @@ mod os {
             ensure_child_stream(&mut child_stdout, StandardStream::Output)?;
             ensure_child_stream(&mut child_stderr, StandardStream::Error)?;
             let cmdline = assemble_cmdline(argv)?;
-            let env_block = config.env.map(format_env_block);
+            let env_block = config.env.map(|env| format_env_block(&env));
             // CreateProcess doesn't search for appname in the PATH.
             // We do it ourselves to match the Unix behavior.
             let executable = config.executable.map(locate_in_path);
@@ -973,7 +1008,7 @@ mod os {
                 executable.as_ref().map(OsString::as_ref),
                 &cmdline,
                 &env_block,
-                &config.cwd.as_ref().map(|os| &os[..]),
+                &config.cwd.as_deref(),
                 true,
                 0,
                 raw(&child_stdin),
@@ -1042,7 +1077,7 @@ mod os {
         }
     }
 
-    fn format_env_block(env: Vec<(OsString, OsString)>) -> Vec<u16> {
+    fn format_env_block(env: &[(OsString, OsString)]) -> Vec<u16> {
         fn to_uppercase(s: &OsStr) -> OsString {
             OsString::from_wide(
                 &s.encode_wide()
@@ -1053,18 +1088,18 @@ mod os {
                             c
                         }
                     })
-                    .collect::<Vec<u16>>(),
+                    .collect::<Vec<_>>(),
             )
         }
-        let mut pruned = {
+        let mut pruned: Vec<_> = {
             let mut seen = HashSet::<OsString>::new();
-            env.into_iter()
+            env.iter()
                 .rev()
-                .filter(|&(ref k, ref _v)| seen.insert(to_uppercase(&k)))
-                .collect::<Vec<_>>()
+                .filter(|&(k, _)| seen.insert(to_uppercase(k)))
+                .collect()
         };
         pruned.reverse();
-        let mut block = Vec::<u16>::new();
+        let mut block = vec![];
         for (k, v) in pruned {
             block.extend(k.encode_wide());
             block.push('=' as u16);
@@ -1075,7 +1110,7 @@ mod os {
         block
     }
 
-    trait PopenOsImpl: super::PopenOs {
+    trait PopenOsImpl {
         fn wait_handle(&mut self, timeout: Option<Duration>) -> io::Result<Option<ExitStatus>>;
     }
 
@@ -1107,7 +1142,7 @@ mod os {
         // failing because the shell tries to redirect stdout to
         // stderr, but fails because it didn't receive a valid stdout.
         if stream.is_none() {
-            *stream = Some(os_common::get_standard_stream(which)?);
+            *stream = Some(get_standard_stream(which)?);
         }
         Ok(())
     }
@@ -1145,7 +1180,7 @@ mod os {
     }
 
     fn assemble_cmdline(argv: Vec<OsString>) -> io::Result<OsString> {
-        let mut cmdline = Vec::<u16>::new();
+        let mut cmdline = vec![];
         let mut is_first = true;
         for arg in argv {
             if !is_first {
@@ -1223,6 +1258,26 @@ impl Drop for Popen {
     }
 }
 
+thread_local! {
+    static STREAMS: RefCell<[Option<Rc<File>>; 3]> = RefCell::default();
+}
+
+#[cfg(unix)]
+use crate::posix::make_standard_stream;
+#[cfg(windows)]
+use crate::win32::make_standard_stream;
+
+fn get_standard_stream(which: StandardStream) -> io::Result<Rc<File>> {
+    STREAMS.with(|streams| {
+        if let Some(ref stream) = streams.borrow()[which as usize] {
+            return Ok(Rc::clone(&stream));
+        }
+        let stream = make_standard_stream(which)?;
+        streams.borrow_mut()[which as usize] = Some(Rc::clone(&stream));
+        Ok(stream)
+    })
+}
+
 /// Error in [`Popen`] calls.
 ///
 /// [`Popen`]: struct.Popen.html
@@ -1230,7 +1285,7 @@ impl Drop for Popen {
 #[derive(Debug)]
 #[non_exhaustive]
 pub enum PopenError {
-    /// The underlying error is io::Error.
+    /// An IO system call failed while executing the requested operation.
     IoError(io::Error),
     /// A logical error was made, e.g. invalid arguments detected at run-time.
     LogicError(&'static str),
@@ -1249,23 +1304,16 @@ impl From<communicate::CommunicateError> for PopenError {
 }
 
 impl Error for PopenError {
-    fn description(&self) -> &str {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
         match *self {
-            PopenError::IoError(ref err) => err.description(),
-            PopenError::LogicError(description) => description,
-        }
-    }
-
-    fn cause(&self) -> Option<&dyn Error> {
-        match *self {
-            PopenError::IoError(ref err) => Some(err as &dyn Error),
-            PopenError::LogicError(_) => None,
+            PopenError::IoError(ref err) => Some(err),
+            PopenError::LogicError(_msg) => None,
         }
     }
 }
 
 impl fmt::Display for PopenError {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         match *self {
             PopenError::IoError(ref err) => fmt::Display::fmt(err, f),
             PopenError::LogicError(desc) => f.write_str(desc),
